from .gamedata import getPlayers, getPointLog, getMatches, getUnplayed
from .pwr import PWRsystems
from .regression import Regression
from .simulate import simulateBracket, simulateMatch, simulateGamelog
from .players import Player, Players
from .tiebreak import getPlayoffSeeding
from .util import playoff_series_ids
from joblib import Parallel, delayed
import pandas as pd
import numpy as np

class Simulate(object):
    def __init__(self, n_sims, pwr_systems=None, rank_adj=0.5, st_dev=1.6, season=2):
        self.n_sims = n_sims
        self.rank_adj = rank_adj
        self.st_dev = st_dev
        self.season = season
        if pwr_systems is None:
            self.pwr_systems = PWRsystems()
        else:
            self.pwr_systems = pwr_systems
        self.players = getPlayers(season)
        self.points = getPointLog(season)
        self.played = getMatches(season)
        self.unplayed = getUnplayed(season)
        for system in self.pwr_systems.systems:
            system.calculate(gamelog=self.points, season=season)
            self.regress(system)
        self.pwr = self.pwr_systems.combine()
        self.regress(self.pwr)

    def run(self, parallel=True, combine=True):
        simulations = []
        if parallel:
            simulations = Parallel(n_jobs=-1)(delayed(self.simulate)() for i in range(self.n_sims))
        else:
            for i in range(self.n_sims):
                simulations.append(self.simulate())
        self.simulations = Simulations(simulations, combine)
        return self

    def playoffs(self, reindex=False):
        if self.simulations.combined:
            return self.copied(self.simulations.playoffs.copy(), reindex)

    def regularseason(self, reindex=False):
        if self.simulations.combined:
            return self.copied(self.simulations.regularseason.copy(), reindex)
    
    def standings(self, reindex=False):
        if self.simulations.combined:
            return self.copied(self.simulations.standings.copy(), reindex)

    def copied(self, df, reindex):
        if reindex:
            return df.reset_index(level='Simulation')
        else:
            return df
            
    def simulate(self):
        return Simulation(self)

    def regress(self, system):
        if system.regress_to is not None:
            if type(system.regress_to) is not Regression:
                system.regress_to = Regression(to=system.regress_to)
            system.regress(system.values)

class Simulation(object):
    def __init__(self, sim):
        self.rankings = sim.pwr.values.copy()
        pwr_adjustments = np.random.normal(0, sim.rank_adj, self.rankings.shape[0])
        self.rankings['PWR'] = self.rankings['PWR'].values - pwr_adjustments
        if sim.unplayed.empty:
            self.regularseason = sim.played
        else:
            simulated = simulateGamelog(sim.unplayed, self.rankings, sim.st_dev, sim.season)
            self.regularseason = pd.concat([sim.played, simulated], ignore_index=True)
        adjusted = pd.DataFrame([x + [1] for x in self.regularseason[['Winner','Loser','W Pts']].values.tolist()] +
                                [x + [0] for x in self.regularseason[['Loser','Winner','L Pts']].values.tolist()],
                                columns=['Player','Opponent','Pts','Wins'])
        df = pd.merge(pd.merge(adjusted, sim.players, on='Player'), 
                      sim.players.rename({'Player':'Opponent','Division':'OppDivision'}, axis=1), on='Opponent')
        self.standings = pd.merge(df.groupby(['Player','Division']).agg({'Pts':'sum','Wins':'sum'}).reset_index(), 
                                  self.rankings, on='Player')
        self.seeding = getPlayoffSeeding(df)
        self.playoffs = self.simulatePlayoffs(sim)
        self.standings = pd.merge(self.standings, self.seeding, how='left', on='Player', suffixes=('', '_')).drop('Division_', axis=1)
        
    def simulatePlayoffs(self, sim):
        players = Players(pd.merge(self.standings, self.seeding, on='Player', suffixes=('', '_')).drop('Division_', axis=1))
        bracket = simulateBracket(players, st_dev=sim.st_dev)
        results = dict((playoff_series_ids[key], value) for (key, value) in bracket.items())
        df = pd.DataFrame.from_dict(results, orient='index').reset_index()
        df[['Round','Series']] = pd.DataFrame(df[['level_0','level_1']].values.tolist(), index=df.index)
        return df[['Round','Series','Winner','Loser','Games']]

class Simulations(object):
    def __init__(self, values, combine=True):
        self.combined = combine
        if combine:
            indices = list(range(len(values)))
            self.playoffs = pd.concat([x.playoffs for x in values], keys=indices).rename_axis(['Simulation','Row'])
            self.standings = pd.concat([x.standings for x in values], keys=indices).rename_axis(['Simulation','Row'])
            self.regularseason = pd.concat([x.regularseason for x in values], keys=indices).rename_axis(['Simulation','Row'])
        else:
            self.values = values



import pandas as pd
import numpy as np

class Regression(object):
    def __init__(self, to=None, weight=None, n_games=33):
        self.regression_values = to
        self.regression_weight = weight
        self.num_games = n_games
    
    def regress(self, df, pwrcol):
        mydf = pd.DataFrame()
        if self.regression_values is None:
            self.regression_values = np.mean(df[pwrcol].values)
        if type(self.regression_values) is not pd.DataFrame:
            reg_values = [self.regression_values] * df.shape[0]
            mydf = df
        else:
            merged = df.merge(self.regression_values, on='Team')
            reg_values = merged['Baseline'].values
            mydf = merged
        if self.regression_weight is not None:
            reg_weight = self.regression_weight * df.shape[0]
            played_weight = (1 - self.regression_weight) * df.shape[0]
            weight_sum = 1
        else:
            reg_weight = np.maximum(self.num_games - mydf['Games Played'].values, [0] * mydf.shape[0])
            played_weight = mydf['Games Played'].values
            weight_sum = reg_weight + played_weight
        played_weighted = mydf[pwrcol].values * played_weight
        regressed_weighted =  reg_values * reg_weight
        return (played_weighted + regressed_weighted) / weight_sum


playoff_series_ids = {
    1:('Round of 16',1),
    2:('Round of 16',2),
    3:('Round of 16',3),
    4:('Round of 16',4),
    5:('Round of 16',5),
    6:('Round of 16',6),
    7:('Round of 16',7),
    8:('Round of 16',8),
    9:('Quarterfinals',1),
    10:('Quarterfinals',2),
    11:('Quarterfinals',3),
    12:('Quarterfinals',4),
    13:('Semifinals',1),
    14:('Semifinals',2),
    15:('Championship',1)}

def extractText(tosearch, delim_left='', delim_right= None,
                reverse_left=False, reverse_right=False,
                optional_left=False, optional_right=False):
    rdelim = delim_left if delim_right is None else delim_right
    returnval = tosearch
    if delim_left != '':
        returnval = extractHelper(returnval, delim_left, reverse_left, optional_left, True)
    if rdelim != '':
        returnval = extractHelper(returnval, rdelim, reverse_right, optional_right, False)
    return returnval

def extractHelper(tosearch, delim, reverse_order, is_optional, is_left):
    cond1 = (is_left == (is_optional == reverse_order))
    cond2 = (not reverse_order) and is_left
    n = 2 if is_left else 0
    basefunc = str.rpartition if reverse_order else str.partition
    partitioned = basefunc(tosearch, delim)
    return partitioned[n] if (cond1 or partitioned[1] == delim) else partitioned[0] if cond2 else ''


from scipy import stats
import numpy as np
import pandas as pd
import re

class PWR(object):
    def __init__(self, weight=1, regress_to=None, values=None):
        self.weight = weight
        self.regress_to = regress_to
        if values is None:
            self.values = None
        else:
            self.values = values.copy()
        
    def calculate(self, **kwargs):
        self.pwrcol = [x for x in list(self.values) if x not in ['Player']][0]
        return self
        
    def regress(self, df):
        self.values[self.pwrcol] = self.regress_to.regress(df, self.pwrcol)

class SRS(PWR):
    def __init__(self, weight=1, regress_to=None):
        PWR.__init__(self, weight, regress_to)
    
    def calculate(self, **kwargs):
        if kwargs['season'] == 1:
            df = kwargs['gamelog'].groupby('Player').agg({'Pts':'mean'})
            df = df.rename(columns={'Pts':'SRS'}).reset_index()
            self.values = df[['Player','SRS']]
        else:
            grouped = kwargs['gamelog'].groupby('Player').agg({'Difference':'sum','Opponent':lambda x: list(x)})
            grouped['Games Played'] = grouped['Opponent'].str.len()
            grouped['Margin'] = grouped['Difference'].values / grouped['Games Played'].values
            grouped['SRS'] = grouped['Margin']
            grouped['OldSRS'] = grouped['Margin']
            players = grouped.to_dict('index')
            for i in range(10000):
                delta = 0.0
                for name, player in players.items():
                    sos = 0.0
                    for opponent in player['Opponent']:
                        sos += players[opponent]['SRS']
                    players[name]['OldSRS'] = player['SRS']
                    players[name]['SRS'] = player['Margin'] + (sos / player['Games Played'])
                    delta = max(delta, abs(players[name]['SRS'] - players[name]['OldSRS']))
                if delta < 0.001:
                    break
            srs_sum = 0.0
            for name, player in players.items():
                srs_sum += players[name]['SRS']
            srs_avg = srs_sum / len(players)
            for name, player in players.items():
                players[name]['SRS'] = player['SRS'] - srs_avg
            df = pd.DataFrame.from_dict(players, orient='index').reset_index()
            self.values = df.rename({'index':'Player'}, axis=1)[['Player','SRS']]
        return self
        
class PWRsystems(object):
    def __init__(self, regress_to=None, srs=None, others=None, scale=None):
        self.regress_to = regress_to
        self.systems = []
        self.scale = None
        if isinstance(scale, dict):
            self.scale = scale
        elif scale:
            self.setDefaultScale()
        if (srs is None) and (others is None):
            self.systems.append(SRS())
        else:
            pairs = [(srs, SRS)]
            for system in [{'Arg':x,'Class':y} for x, y in pairs]:
                if type(system['Arg']) is bool:
                    if system['Arg']:
                        self.systems.append(system['Class']())
                elif system['Arg'] is not None:
                    self.systems.append(system['Arg'])
            if others is not None:
                if isinstance(others, PWR):
                    self.systems.append(others)
                else:
                    for system in others:
                        self.systems.append(system)
                        
    def setDefaultScale(self):
        self.scale = {'st_dev':1.05,'mean':0}

    def combine(self):
        if (len(self.systems) > 1) and (self.scale is None):
            self.setDefaultScale()
        self.combined = self.systems[0].values[['Player']]
        for system in self.systems:
            self.combined = pd.merge(self.combined, system.values, on='Player', suffixes=('','_'))
            new_z = stats.zscore(self.combined[system.pwr].values)
            new_weights = [system.weight] * self.combined.shape[0]
            if 'z_scores' not in self.combined:    
                self.combined['z_scores'] = [[x] for x in new_z]
                self.combined['weights'] = [[x] for x in new_weights]
            else:
                self.combined['z_scores'] = [x[0] + [x[1]] for x in list(zip(self.combined['z_scores'].values, new_z))]
                self.combined['weights'] = [x[0] + [x[1]] for x in list(zip(self.combined['weights'].values, new_weights))]
        zipped = zip(self.combined['z_scores'].values, self.combined['weights'].values)
        self.combined['Avg_z'] = [np.inner(x, y) / np.sum(y) for x, y in zipped]
        if self.scale is None:
            self.combined['PWR'] = self.combined[self.systems[0].pwr].values
        else:
            self.combined['PWR'] = self.combined['Avg_z'].values * self.scale['st_dev'] + self.scale['mean']
        return PWR(regress_to=self.regress_to, values=self.combined[['Player','PWR']], pwr='PWR')


from .players import Player, Players
from scipy import stats
import pandas as pd
import numpy as np

#simulates games in a gamelog using standard method
def simulateGamelog(gamelog, rankings, st_dev, season):
    merged = pd.merge(gamelog, rankings.rename({'Player':'Home'}, axis=1), on='Home').rename({'PWR':'Home PWR'}, axis=1)
    merged = pd.merge(merged, rankings.rename({'Player':'Away'}, axis=1), on='Away').rename({'PWR':'Away PWR'}, axis=1)
    dist = stats.norm(merged['Home PWR'].values - merged['Away PWR'].values, st_dev)
    test_vals = 1 - dist.cdf(0)
    random_vals = np.random.random((3, merged.shape[0]))
    w_pts = 4 if season == 1 else 3
    home_win = np.sum(np.less(random_vals, test_vals).astype(int), axis=0)
    merged['Winner'] = np.where(np.greater(home_win, 1), merged['Home'], merged['Away'])
    merged['Loser'] = np.where(np.greater(home_win, 1), merged['Away'], merged['Home'])
    merged['W Pts'] = np.where(np.isin(home_win, [1, 2]), w_pts - 1, w_pts)
    merged['L Pts'] = w_pts - merged['W Pts'].values
    return merged[['Winner','Loser','W Pts','L Pts']]

#simulates a "bracket"-style group of games between seeded teams
def simulateBracket(players, st_dev, n_winners=1):
    def getWinners(matchups, results=None, round_n=None, seriescounter=None):
        def roundCount(matchups):
            return 1 + (0 if not isinstance(matchups[0], list) else roundCount(matchups[0]))
        original_call = False
        if round_n is None:
            original_call = True
            results = []
            round_n = roundCount(matchups)
            seriescounter = {x:0 for x in range(round_n + 1)}
        if isinstance(matchups[0], list):
            round_n -= 1
            winner_a = getWinners(matchups[0], results, round_n, seriescounter)
            winner_b = getWinners(matchups[1], results, round_n, seriescounter)
            winner = getWinners([winner_a, winner_b], results, round_n + 1, seriescounter)
            if original_call:
                final_results = {}
                cumulative_series_count = {-1:0}
                for i, x in seriescounter.items():
                    cumulative_series_count[i] = cumulative_series_count[i - 1] + x
                for i, x in enumerate(results):
                    result = {'Winner':x['Winner'],'Loser':x['Loser'],'Games':x['Games']}
                    final_results[cumulative_series_count[x['Round'] - 1] + x['Series']] = result
                return final_results
            else:
                return winner
        else:
            n_games = 5 if round_n < 3 else 7
            result = simulateMatch(matchups[0], matchups[1], st_dev=st_dev, n_games=n_games)
            seriescounter[round_n] = seriescounter[round_n] + 1
            results.append({'Winner':result['Winner'].name,'Loser':result['Loser'].name,
                            'Games':result['Games'],'Round':round_n,'Series':seriescounter[round_n]})
            return result['Winner']
    n_players = players.len()
    remaining = players
    results = {}
    remaining.index(seed=True)
    remaining_seeds = [x[0] for i, x in remaining.items()]
    seeds = remaining_seeds
    while True:
        seeds = [[seeds[i], seeds[-i-1]] for i in range(int(len(seeds) / 2))]
        if len(seeds) == 2:
            break
    return getWinners(seeds)

def simulateMatch(player_a, player_b, st_dev, n_games):
    target_wins = int(n_games / 2) + 1
    pwr_difference = [player_a.pwr - player_b.pwr] * n_games
    win_probability = 1 - stats.norm(pwr_difference, st_dev).cdf(0)
    is_winner = (np.random.random(n_games) < win_probability).astype(int)
    a_wins = 0
    b_wins = 0
    for i in range(n_games):
        a_wins += is_winner[i]
        b_wins += 1 - is_winner[i]
        if a_wins == target_wins:
            return {'Winner':player_a,'Loser':player_b,'Games':i + 1}
        elif b_wins == target_wins:
            return {'Winner':player_b,'Loser':player_a,'Games':i + 1}


from requests import get
from io import StringIO
import numpy as np
import pandas as pd

def getPlayers(season=None):
    df = pd.read_csv(StringIO(get('https://pastebin.com/raw/WEf5d2fm').text))
    if season is None:
        season = df['Season'].max()
    return df[df['Season'] == season][['Player','Division']]

def getMatches(season=None):
    games = pd.read_csv(StringIO(get('https://pastebin.com/raw/eTmbZM46').text))
    if season is None:
        season = games['Season'].max()
    games = games[games['Season'] == season]
    games_grouped = games.groupby(['MatchID','Division','Week','Match'])
    lm = lambda x:max(set(list(x)), key=list(x).count)
    matches = pd.merge(games_grouped.agg({'Winner':lm,'Loser':lm}), 
                       games_grouped.agg(lambda x: tuple(x)), 
                       left_index=True, right_index=True, suffixes=('','_list')).reset_index()
    #calculate penalties for mid-match forfeits
    lm = lambda x:sum(np.where(np.repeat(x[0], len(x[1])) == np.array(x[1]), np.isnan(x[2]), 0))
    w_pts = 4 if season == 1 else 3
    w_penalty = [lm(x) * 2 for x in matches[['Loser','Winner_list','L Clears']].values]
    l_penalty = [(lm(x) in [1, 2]) * 2 for x in matches[['Loser','Loser_list','L Clears']].values]
    no_penalty = [sum(np.where(np.isnan(x), 1, 0)) == 3 for x in matches['W Clears'].values]
    #calculate match point totals
    is_sweep = [~np.isin(x[0], x[1]) for x in matches[['Loser','Winner_list']].values]
    matches['W Pts'] = np.where(is_sweep, w_pts, w_pts - 1) - np.where(no_penalty, 0, w_penalty)
    matches['L Pts'] = np.where(is_sweep, 0, 1) - np.where(no_penalty, 0, l_penalty)
    repl_df = pd.read_csv(StringIO(get('https://pastebin.com/raw/0kyaPRqX').text))
    replacements = repl_df.to_dict('records')
    for repl in replacements:
        matches['Winner'] = matches['Winner'].str.replace(repl['Old'], repl['New'])
        matches['Loser'] = matches['Loser'].str.replace(repl['Old'], repl['New'])
    return matches[['Winner','Loser','W Pts','L Pts']]
    
def getPointLog(season=None):
    games = pd.read_csv(StringIO(get('https://pastebin.com/raw/eTmbZM46').text))
    if season is None:
        season = games['Season'].max()
    games = games[games['Season'] == season]
    if season == 1:
        points = pd.DataFrame(games[['Winner','Loser','W Clears','W GmOv','W TB']].values.tolist() +
                              games[['Loser','Winner','L Clears','L GmOv','L TB']].values.tolist(), 
                              columns= ['Player','Opponent','Clears','GmOv','TB'])
        points = points[~np.isnan(points['Clears'])]
        points['Pts'] = points['Clears'] - points['GmOv']
        return points
    else:
        points = pd.DataFrame(games[['Winner','Loser','W Clears','W TB','L Clears','L TB']].values.tolist() +
                              games[['Loser','Winner','L Clears','L TB','W Clears','W TB']].values.tolist(), 
                              columns= ['Player','Opponent','Clears','TB','OppClears','OppTB'])
        league_avg = points['Clears'].mean()
        clears = points.copy()
        clears['Opponent'] = 'League'
        clears['OppClears'] = league_avg
        clears['OppTB'] = 0
        clears_dup = pd.DataFrame(clears[['Player','Opponent','Clears','TB','OppClears','OppTB']].values.tolist() +
                                  clears[['Opponent','Player','OppClears','OppTB','Clears','TB']].values.tolist(), 
                                  columns= ['Player','Opponent','Clears','TB','OppClears','OppTB'])
        full_df = pd.concat([points, clears_dup])
        full_df = full_df[~np.isnan(full_df['Clears'])]
        full_df = full_df[~np.isnan(full_df['OppClears'])]
        full_df['Difference'] = full_df['Clears'] - full_df['OppClears']
        return full_df

def getUnplayed(season=None):
    df = pd.read_csv(StringIO(get('https://pastebin.com/raw/rmb6A8uW').text))
    if season is None:
        season = df['Season'].max()
    return df[df['Season'] == season][['Home','Away']]


import numpy as np
import pandas as pd

def getPlayoffSeeding(gamelog):
    seeds = getSeeds(gamelog, [1,2,3,4,5,6,7,8], list(set(gamelog['Player'].values)))
    for i, seed in enumerate(seeds):
        seeds[i]['Seed'] = seed['Seed'] * 2 - (1 if seed['Division'] == 'B' else 0)
    return pd.DataFrame(seeds)

def getSeeds(gamelog, seeds, winners):
    wins = gamelog.groupby(['Player','Division'])['Wins'].sum()
    seeding = []
    for div, g in wins.groupby('Division'):
        filtered = g[[x in winners for x in g.index.get_level_values('Player').values]]
        w_rank = filtered.rank(method='min', ascending=False).values
        for s, seed in enumerate(seeds):
            iv = filtered[w_rank <= s + 1].index.values
            indices = [x for x in iv if x[0] not in [x['Player'] for x in seeding]]
            winner = breakTies(gamelog, [{'Player':t} for t, d in indices])
            seeding.append({'Division':div,'Player':winner,'Seed':seed})
    return seeding

def breakTies(gamelog, tiedplayers):
    if len(tiedplayers) == 1:
        return tiedplayers[0]['Player']
    remainder = [x['Player'] for x in tiedplayers]
    filtered = gamelog[gamelog['Player'].isin(remainder)]
    grouped = filtered.groupby(['Player'])['Pts'].sum().reset_index()
    remainder = grouped[(grouped['Pts'].values == grouped['Pts'].values.max())]['Player'].values
    if len(remainder) == 1:
        return remainder[0]
    elif len(remainder) == 2 and len(tiedplayers) != 2:
        return breakTies(gamelog, [x for x in tiedplayers if x['Player'] in remainder])
    filtered = gamelog[gamelog['Player'].isin(remainder) & gamelog['Opponent'].isin(remainder)]
    if not filtered.empty:
        grouped = filtered.groupby(['Player']).agg({'Pts':'sum'}).reset_index()
        remainder = grouped[(grouped['Pts'].values == grouped['Pts'].values.max())]['Player'].values
    if len(remainder) == 1:
        return remainder[0]
    elif len(remainder) == 2 and len(tiedplayers) != 2:
        return breakTies(gamelog, [x for x in tiedplayers if x['Player'] in remainder])
    return np.random.choice(remainder)


import pandas as pd
from itertools import compress
from collections import OrderedDict

class Player(object):
    def __init__(self, name, div=None, pwr=None, seed=None):
        self.name = name
        self.division = div
        self.pwr = pwr
        self.seed = seed

class Players(object):
    def __init__(self, values, indexed=False):
        if type(values) is pd.DataFrame:
            playerdict = values.to_dict('records')
            self.values = [Player(name=x['Player'], div=x['Division'], 
                                  pwr=x['PWR'], seed=x['Seed']) for x in playerdict]
        else:
            self.values = values
        self.indexed = indexed

    def index(self, name=False, div=False, seed=False):
        attrs = ['name','division','seed']
        bools = [name, div, seed]
        indices = list(compress(attrs, bools))
        multi_index = len(indices) != 1
        val_dict = {}
        for player in self.values:
            if multi_index:
                key = tuple([getattr(player, x) for x in indices])
            else:
                key = getattr(player, indices[0])
            if key in val_dict:
                val_dict[key] = val_dict[key] + [player]
            else:
                val_dict[key] = [player]
        self.values = OrderedDict(sorted(val_dict.items(), key=lambda x: x[0]))
        self.indexed = True
        return self

    def copy(self, reset_index=True):
        if not reset_index:
            return Players(self.values, indexed=self.indexed)
        elif self.indexed:
            return Players([x for y in [x for i, x in self.items()] for x in y])
        else:
            return Players(self.values)            

    def len(self):
        return len(self.values)

    def items(self):
        return self.values.items()

    def keys(self):
        if self.indexed:
            return list(self.values)



from setuptools import setup

setup(
    name='smm2sim',
    url='https://github.com/dmparker0/smm2sim/',
    author='Dan Parker',
    author_email='dan.m.parker0@gmail.com',
    packages=['smm2sim'],
    install_requires=['numpy','pandas','scipy','bs4','requests','joblib'],
    version='1.0.6',
    license='MIT',
    description='A tool for simulating the GSA Mario Maker 2 Endless Expert League',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    download_url = 'https://github.com/dmparker0/smm2sim/archive/v1.0.6.tar.gz',
    keywords = ['Mario','SMM','SMM2','GSA','speedrun','simulation','statistics'], 
)
